"""Module containing get_info and wavread programs for a Broadcast Wave
file written by an ZOOM F8 recorder.
"""
from __future__ import division

def get_info(file, info={}):
    """Read the information in a Broadcast Wave Format (BWF) file
    written by the ZOOM F8 recorder and return the contents.
    """
    """info = get_info(file, info={}) Read the information in a WAV file
    written by a Zoom device (this function was written specifically for
    recordings made by the Zoom F8 recorder but it may work with other
    Zoon devices as well) and return the contents. The standard
    information in the fmt chunk is included in the info dictionary as
    well as information in the bext chunk and the iXML chunk.
    Input Parameters
    
        file - filehandle of an open WAV file
        info - (optional) dictionary that may contain file
            information from other sources. Defaults to an empty
            dictionary.
    
    Output
    
        info - dictionary with information read from the file. If an
            info dictionary was supplied as an input parameter,
            entires that were not changed are also included.
    
    Standard info dictionary keys and values returned
    
        "bits" - integer with the number of bits in each sample.
        "block_align" - number of bytes sampled at the same time (all
            channels combined) in the data
        "byte_per_s" - integer number of bytes per second recorded
        "chan" - integer number of channels in the file
        "compress" - integer Wave file compression index. Only 1
            (uncompressed integer data) and 3 (uncompressed floating
            point data) are currently supported.
        "data0" - integer byte address of the first sample in the file
        "filesize" - integer size of the file in bytes
        "fs" - integer sample rate in samples/second
        "Nsamples" - integer number of samples in the file (in each
            channel)
        "wavetype" - string with "decimus" as the wave file type read.
    
    bext Chunk info dictionary keys and values returned. (See Zoom
    documentation for details.)
    
        "CodingHistory" - coding history string
        "desc" - recording description string
        "LoudnessRange" - int16 recording loudness range value
        "LoudnessValue" - int16 recording loudness value
        "MaxMomentaryLoudness" - int16 recording maximum momentary
            loudness value
        "MaxShortTermLoudness" - int16 recording maximum short term
            loudness value
        "MaxTruePeakLevel" - int16 recording maximum maximum true
            peak level
        "OriginationDate" - recording origination date string
        "OriginationTime" - recording origination time string
        "Originator" - recording originator string
        "OriginatorReference" - recording originator reference string
        "TimeReferenceHigh" - int32 time of high sample in recording
        "TimeReferenceLow" - int32 time of low sample in recording
        "UMID" - UMID string
        "Version" - int16 version number
        
    The contents in the entire iXML block are stored in info["iXML"] as
    a string."""

    import struct
    import numpy as np
    file.seek(0)
    data = file.read(4)
    if data != b'RIFF':
        raise ValueError('Chunk ID not RIFF: ', data)
    data = file.read(4)
    info['filesize'] = struct.unpack('<I',data)[0] + 8
    data = file.read(4)
    if data != b'WAVE':
        raise ValueError('Chunk ID not WAVE: ', data)
    data = file.read(4)
    if data != b'bext':
        raise ValueError('Chunk ID not bext: ', data)
    bext_size = struct.unpack('<I', file.read(4))[0]
    data = file.read(256)  
    info['desc'] = data.split(b'\0',1)[0].decode()
    data = file.read(32)
    info['Originator'] = data.split(b'\0',1)[0].decode()
    data = file.read(32)
    info['OriginatorReference'] = data.split(b'\0',1)[0].decode()
    info['OriginationDate'] = file.read(10).decode()
    info['OriginationTime'] = file.read(8).decode()
    info['TimeReferenceLow'] = struct.unpack('<I', file.read(4))[0]
    info['TimeReferenceHigh'] = struct.unpack('<I', file.read(4))[0]
    info['Version'] = struct.unpack('<H',file.read(2))[0]
    info['UMID'] = file.read(64).decode()
    info['LoudnessValue'] = struct.unpack('<H',file.read(2))[0]
    info['LoudnessRange'] = struct.unpack('<H',file.read(2))[0]
    info['MaxTruePeakLevel'] = struct.unpack('<H',file.read(2))[0]
    info['MaxMomentaryLoudness'] = struct.unpack('<H',file.read(2))[0]
    info['MaxShortTermLoudness'] = struct.unpack('<H',file.read(2))[0]
    file.seek(180,1)
    data = file.read(bext_size - 602)
    info['CodingHistory'] = data.split(b'\0',1)[0].decode()
    data = file.read(4)
    if data == b'iXML':
        iXML_size = struct.unpack('<I', file.read(4))[0]
        data = file.read(iXML_size)
        info['iXML'] = data.split(b'\0',1)[0].decode()
        data = file.read(4)
    if data != b'fmt ':
        raise ValueError('Chunk ID not fmt : ', data)
    fmt_size = struct.unpack('<I', file.read(4))[0]
    info['compress'] = struct.unpack('<H', file.read(2))[0]
    info['chan'] = struct.unpack('<H', file.read(2))[0]
    info['fs'] = struct.unpack('<I', file.read(4))[0]
    info['byte_per_s'] = struct.unpack('<I', file.read(4))[0]
    info['block_align'] = struct.unpack('<H', file.read(2))[0]
    info['bits'] = struct.unpack('<H', file.read(2))[0]
    if fmt_size > 16:
        size = struct.unpack('<H', file.read(2))[0]
        info['extra fmt'] = file.read(size)
    data = file.read(4)
    if data == b'PAD ':
        pad_size = struct.unpack('<I', file.read(4))[0]
        file.seek(pad_size,1)
        data = file.read(4)
    if data != b'data':
        raise ValueError('Chunk ID not data: ', data)
    # We are at the data chunk.
    # Nsamples is the number samples in the file (should be total time * fs).
    info['Nsamples'] = struct.unpack('<I', file.read(4))[0] // \
        (info['block_align'])
    # data0 is the position (in bytes) of the first data sample.
    info['data0'] = file.tell()
    info['wavetype'] = 'zoom'
    return(info)


from __future__ import division
import numpy as np
from scipy.signal import bilinear

def spl(data, fs, weighting="A", tconst=0.125, pref=20.0, cal=1.0, pms=0., 
    pms_ret=False):
    """Return an array of sampled sound pressure levels using time constant T.  
    Usage
    
        SPL = spl(data, fs, weighting='A', tconst=0.125, pref=20.0)
    
    Input Parameters
    
    data: an array of sampled sound pressures.
    fs: sampling frequency in hertz.
    weighting: type of weighting to use. This parameter can be a
        string to represent preset values 'A' for A-weighting, 'M'
        for M-weighting (see documentation on the function weight()
        to set frequency parameters). It can also be a function that
        provides digital filter parameters to the weight() function.
        For no weighting, use weighting = 1. The default is 'A'
        weighting.
    tconst: time constant. Defaullts to 0.125 s (fast). This parameter can be 
        the value in seconds or preset values given with the strings 'Fast' 
        (0.125 s), 'Slow' (1.000 s), or "Impulse' (0.035 s).
    pref: reference pressure. Defaults to 20.0 (micropascals, standard for
        atmospheric sounds). Use 1.0 (micropascals) for underwater sounds.
    cal: calibration factor of the recording. This is the value that
        converts data samples to the appropriate pressure units
        (micropascals). The default value is 1 (no calibration
        adjustment).
    pms: an initial value for the mean square pressure 'historical'
        value for time constant. Use this to continue the calculation
        from another recording. Defaults to 0.0.
    pms_return: whether or not to return the mean square pressure value for 
        subsequent calculations. Defaults to False.
    
    Output
    
    SPL: a numpy array of sampled sound pressure levels corresponding to the 
        the same sampling times a the elements of data.  Note that the initial 
        elements SPL[i] are based on a truncated history because they only use 
        pressure values from data[i] back to data[0].
    pms: The mean square sound pressure for use in subsequent
        calculations such as the recording continuing in another
        file. Only returned if the input parameter pms_return is True.
    """
    if tconst == 'Fast':
        tconst = 0.125
    elif tconst == 'Slow':
        tconst = 1.0
    elif tconst == 'Impulse':
        tconst = 0.035
    if weighting == 'M':
        weighting = M_weighting
    elif weighting == 'A':
        weighting = A_weighting
    if hasattr(weighting, '__call__'):
        wdata = weight(data, weighting, fs)
    else:
        wdata = data
    caldB = 20. * np.log10(cal)
    spldata = np.zeros(wdata.size)
    for n in range(wdata.size):    
        pms = pms * np.exp(-1 / np.float64(tconst * fs)) + wdata[n]**2 / \
            np.float64(tconst * fs)
        # Calibrate the SPL (better precision with lower numbers).
        spldata[n] = (10.0 * np.log10(pms/pref**2)  + caldB)
    if pms_ret:
        return(spldata, pms)
    else:
        return(spldata)
    
def sel(data, fs, weighting="A", pref=20.0, cal=1.0, amb=-1.0e3, pms=0., \
    pms_ret=False):
    """Return the sound exposure level of a sampled sound.  
    Usage
    
        SEL = sel(data, fs, weighting='A', pref=20.0, cal=1.0,
                  amb=-1.0e3, pms=0., pms_ret=False))
    
    Input Parameters
    
    data: an array of sampled sound pressures.
    fs: sampling frequency in hertz.
    weighting: type of weighting to use. This parameter can be a
        string to represent preset values 'A' for A-weighting, 'M'
        for M-weighting (see documentation on the function
        SPLWeighting.weight() to set frequency parameters). The
        parameter weighting can also be a function that provides
        digital filter parameters to the SPLWeighting.weight()
        function. For no weighting, use weighting=1. The default is
        'A' weighting.
    pref: reference pressure. Defaults to 20.0 (micropascals, the
        default value for atmospheric sounds). Use 1.0 (micropascals)
        for underwater sounds.
    cal: calibration factor of the recording. This is the value that converts 
        data samples to the appropriate pressure units (micro Pa).
    pms: an initial value for the mean square pressure 'historical' value for 
        time constant. Defaults to 0.0.
    pms_return: whether or not to return the mean square pressure
        value (in linear units) for subsequent calculations.  This
        allows the SEL to be calculated from several sound
        recordings. Defaults to False.
    
    Output
    
    SEL: the sound exposure level of the recording in dB re pref.
        Note that if pms_return is True, the intermediate value of
        mean square pressure is returned in linear units (see
        pms_return above). 
    """
    p2amb = pref**2 * 10.0**(amb/10.0)
    if weighting == 'M':
        weighting = SPLWeighting.M_weighting
    elif weighting == 'A':
        weighting = SPLWeighting.A_weighting
    if hasattr(weighting, '__call__'):
        wdata = SPLWeighting.weight(data, weighting, fs)
    else:
        wdata = data
    for p in wdata:
        pms = pms + ((p * cal)**2 - p2amb) / fs
    if pms_ret:
        return(pms)
    else:
        selcal = 10.0 * np.log10(pms / pref**2)
        return(selcal)
        
def spl_wav(wavfile, t0=0, t1=-1, weighting='A', tconst=0.125, pref=20.0, \
    cal=1.0, outfile=True, output=False):
    """Return sampled sound pressure levels from data in a WAV file.
    Usage
    
        spl_wav(wavfile, t0=0, t1=-1, weighting='A', tconst=0.125,
            pref=20.0, outfile=True, output=False)
    
    Input Parameters
    
    wavfile: filename of the input WAV file.
    t0: Start time in seconds for data to be analyzed in the WAV file. Defaults 
        to 0.
    t1: End time in seconds for data to be analyzed in the WAV file. A value of 
        -1 (default) represents the end of the file.
    weighting: type of weighting to use. A value of 1 represents no
        weighting. The default value is 'A' for A-weighting. See
        documentation for the function spl().
    tconst: time constant. Defaullts to 0.125 s (fast). This parameter can be 
        the value in seconds or preset values given with the strings 'Fast' 
        (0.125 s), 'Slow' (1.000 s), or "Impulse' (0.035 s).
    pref: reference pressure to normalize output. This should be the numerical 
        value in data is 0 dB. Defaults to 20.0 (micropascal).
    cal: calibration factor of the recording. This is the value that converts 
        data samples to the appropriate pressure units (micro Pa).
    outfile: output file name. A value of True (default) uses the basename of 
        the input file with '_SPL' appended and the extension '.csv'. A string 
        value provides a filename. A value of False results in no output file 
        written.
    output: boolean that determines whether or not the SPL values are returned 
        as output.  A value of True produces output. Defaults to False.
    
    Output
    
    Output is produced only if the parameter output is True.  In this case 
    the output parameters are time, SPL.
        
    time: time values in seconds corresponding SPL samples. The time values are 
        the same as the values in the input WAV file. 
    SPL: a numpy array of sampled sound pressure levels corresponding to the 
        the same sampling times a the elements of data.  Note that the initial 
        elements SPL[i] are based on a truncated history because they only use 
        pressure values from data[i] back to data[0].
    
     The output file is a CSV file with rows of (time, SPL) pairs. 
    """
    from os.path import splitext
    import wavefile
    import csv
    fs, data = wavfile.read(wavfile, t0, t1)
    SPL = spl(data, fs, weighting, tconst, pref, cal)
    time = np.divide(np.arange(SPL.size, dtype=np.float64), np.float64(fs))
    time = np.add(time, np.float64(t0))
    if outfile:
        if outfile == True:
            base, _ = splitext(wavfile)
            outfile = base + '_SPL.csv'
        with open(outfile, 'wb') as csvfile:
            csvwrite = csv.writer(csvfile)
            csvwrite.writerow(['Time (s)', 'SPL (dB)'])
            for n in range(SPL.size):
                csvwrite.writerow([time[n], SPL[n]])
    if output:
        return(time, SPL)

def spl_wav_dir(pattern='*.wav', outname='_SPL.csv', weighting='A.', \
    tconst=0.125, pref=20.0, cal=1.0, icL=False, verbose=False):
    """
    Evaluate all files matching pattern using spl_wav() and write output to 
    files.
    
    Input Parameters
    
    pattern: UNIX-style pathname pattern for input files. See the glob module 
        for details. Defaults to '*.wav', which returns all files in the current 
        directory with extension .wav.
    outname: string appended to the basename of the input file to give the 
        output file name. Defaults to '_SPL.csv'
    weighting: type of weighting to use. A value of 1 represents no
        weighting. The default value is 'A' for A-weighting. See
        documentation for the function spl().
    tconst: time constant. Defaullts to 0.125 s (fast). This parameter can be 
        the value in seconds or preset values given with the strings 'Fast' 
        (0.125 s), 'Slow' (1.000 s), or "Impulse' (0.035 s).
    pref: reference pressure to normalize output. This should be the numerical 
        value in data is 0 dB. Defaults to 20.0 (micropascal).
    cal: calibration factor of the recording. This is the value that converts 
        data samples to the appropriate pressure units (micro Pa).
    icL: boolean that determines if input files were recorded by an icListen 
        hydrophone. If so, the calibration information is extracted from the 
        information embedded in the file. Defaults to False.
    verbose: print progress message. Defaults to False.
    """
    from glob import glob
    from os.path import splitext
    from .recorders.icListen import get_info
    files = glob(pattern)
    fcount = 0
    for file in files:
        if verbose:
            print('Processing file ' + file + ' ...')
        base, _ = splitext(file)
        outfile = base + outname
        if icL:
            cal = get_info(file)['cal']
            if verbose:
                print('Found icListen info. cal: ' + cal)
        spl_wav(file, weighting=weighting, tconst=tconst, pref=pref, \
            cal=cal, outfile=outfile)
        if verbose:
            print('Output written to file ' + outfile)
            fcount += 1
    if verbose:
        print('Complete... Processed ' + str(fcount) + ' files.')

def spl_wav_files(infnames, outname='_SPL.csv', weighting='A', \
    tconst=0.125, pref=20.0, cal=1.0, icL=False, verbose=False):
    """
    Evaluate all files in infnames matching using spl_wav() and write output to 
    files.
    
    Input Parameters
    
    infnames: list of the input files to be processed.
    outname: string appended to the basename of the input file to give the 
        output file name. Defaults to '_SPL.csv'
    weighting: type of weighting to use. A value of 1 represents no
        weighting. The default value is 'A' for A-weighting. See
        documentation for the function spl().
    tconst: time constant. Defaullts to 0.125 s (fast). This parameter can be 
        the value in seconds or preset values given with the strings 'Fast' 
        (0.125 s), 'Slow' (1.000 s), or "Impulse' (0.035 s).
    pref: reference pressure to normalize output. This should be the numerical 
        value in data is 0 dB. Defaults to 20.0 (micropascal).
    cal: calibration factor of the recording. This is the value that converts 
        data samples to the appropriate pressure units (micro Pa).
    icL: boolean that determines if input files were recorded by an icListen 
        hydrophone. If so, the calibration information is extracted from the 
        information embedded in the file. Defaults to False.
    verbose: print progress message. Defaults to False.
    """
    from os.path import splitext
    from .recorders.icListen import get_info
    fcount = 0
    for file in infnames:
        if verbose:
            print('Processing file ' + file + ' ...')
        base, _ = splitext(file)
        outfile = base + outname
        if icL:
            cal = get_info(file)['cal']
            if verbose:
                print('Found icListen info. cal: ' + cal)
        spl_wav(file, weighting=weighting, tconst=tconst, pref=pref, \
            cal=cal, outfile=outfile)
        if verbose:
            print('Output written to file ' + outfile)
            fcount += 1
    if verbose:
        print('Complete... Processed ' + str(fcount) + ' files.')


def A_weighting(fs):
    """Design of an A-weighting filter.
    b, a = A_weighting(fs) designs a digital A-weighting filter for
    sampling frequency `fs`. Usage: y = scipy.signal.lfilter(b, a, x).
    Warning: `fs` should normally be higher than 20 kHz. For example,
    fs = 48000 yields a class 1-compliant filter.
    References:
       [1] IEC/CD 1672: Electroacoustics-Sound Level Meters, Nov. 1996.
    
    Translated from a MATLAB script (which also includes C-weighting,
    octave and one-third-octave digital filters). Author: Christophe
    Couvreur, Faculte Polytechnique de Mons (Belgium)
        couvreur@thor.fpms.ac.be
    Last modification: Aug. 20, 1997, 10:00am. BSD license
    http://www.mathworks.com/matlabcentral/fileexchange/69 Translated
    from adsgn.m to Python 2009-07-14 endolith@gmail.com
    """
    # Definition of analog A-weighting filter according to IEC/CD 1672.
    f1 = 20.598997
    f2 = 107.65265
    f3 = 737.86223
    f4 = 12194.217
    A1000 = 1.9997

    NUMs = [(2*np.pi * f4)**2 * (10**(A1000/20)), 0, 0, 0, 0]
    DENs = np.polymul([1, 4*np.pi * f4, (2*np.pi * f4)**2],
                   [1, 4*np.pi * f1, (2*np.pi * f1)**2])
    DENs = np.polymul(np.polymul(DENs, [1, 2*np.pi * f3]),
                                 [1, 2*np.pi * f2])

    # Use the bilinear transformation to get the digital filter.
    # (Octave, MATLAB, and PyLab disagree about Fs vs 1/Fs)
    return bilinear(NUMs, DENs, fs)

def M_weighting(fs, flow=40.0, fhigh=300.0):
    """Design of an M-weighting filter for marine bioacoustics 
    b, a = M_weighting(fs, flow=40.0, fhigh=300.0) designs a digital
    M-weighting filter for sampling frequency `fs`. 
    Usage: y = scipy.signal.lfilter(b, a, x).
    Warning: `fs` should normally be higher than 20 kHz.
    Reference:
      Southall, B. L., Bowles, A. E., Ellison, W. T., Finneran, J. J., Gentry, 
      R. L., Greene Jr, C. R., Kastak, D., Ketten, D. R., Miller, J. H., 
      Nachtigall, P. E., Richardson, W. J., Thomas, J. A., and Tyack,
      P. L. (2007), "Marine mammal noise exposure criteria: Initial scientific 
      recommendations," Aquatic Mammals 33(4), 411--521.
    """
    NUMs = [(2*np.pi * fhigh)**2 + (2*np.pi * flow)**2, 0, 0]
    DENs = np.polymul([1, 4*np.pi * fhigh, (2*np.pi * fhigh)**2], \
        [1, 4*np.pi * flow, (2*np.pi * flow)**2])
    return bilinear(NUMs, DENs, fs)

def weight(data, weighting, fs):
    """Calls the function weighting(fs) to generate a digital filter and then 
    filters data by that filter.  fs is the sampling frequency of the data.
    Usage:
        wdata = weight(data, weighting, fs)
    To provide extra parameters to weighting, use the lambda function. For 
    example:
        wdata = weight(data, lambda fs: M_weighting(fs, flow=150.0, 
                fhigh=1.6e5), fs)
    """
    from scipy.signal import lfilter
    b, a = weighting(fs)
    return(lfilter(b, a, data))



"""Read and write data to/from Broadcast Wave files.  The following
propriatory boadcast wave file formats are currently supported:

  generic - generic Windows WAVE file format containing the basic
       infromation in the WAVE format chunk. No additional metadata is read.
       
  decimus - Windows WAVE file written by the Decimus(R) passive
      acoustics monitoring system and other devices that use the SA
      Instrumentation DAQ card. Metadata is extracted from the filename,
      which includes a timestamp, into the info dictionary.

  icListen - WAVE files written by icListen(R) recording devices.
      Metadata in the INFO chunk is read into the info dictionary.

  zoom - WAVE files written by ZOOM(R) recording devices. Metadata
      from the bext chunk and the iXML chunk is read into the info
      dictionary.
"""

from __future__ import division, print_function

def read(filename, t0=0, t1=-1, wavetype=None, chunk_b=3072, verbose=False):
    """info, wave = read(filename, t0, t1, wavetype, chunk_b)
    
    Read a WAV file and return the file information and waveform data.
    Input parameters
    
      filename - string with the name of the input WAV file
      t0 - start time in seconds for returned data (default: 0)
      t1 - end time in seconds for returned data. Value of -1 represents
        the end of the file. (default: -1)
      wavetype - string representing the type of WAV file (default:
        None). Currnetly supported types are 'generic', 'decimus',
        'icListen', and 'zoom'. If the value is None, the wavetype
        is determined using identify.
      chunk_b - number of bytes for each data chunk read from the
        file (default: 3072)
      verbose - give verbose status updates (default: False)
     
    Output
    
      info - dictionary with file information
      wave - Numpy array with waveform data values
    """
    import struct
    import numpy as np
    file = open(filename, 'rb')
    info = getInfo(file, wavetype)
    wave = wave_chunk(file, info, t0, t1, chunk_b=chunk_b, verbose=verbose)
    file.close()
    return(info, wave)

def wavread(filename, t0=0, t1=-1, wavetype=None, chunk_b=3072):
    """Depricated: This is a frontend to wavefile.read maintained for
    legacy purposes. See wavefile.read for complete usage information.
    """
    import warnings
    deprication_msg = "wavefile.wavread() is depricated. Use wavefile.read()."
    warnings.warn(deprication_msg, FutureWarning)
    info, wave = read(filename, t0, t1, wavetype, chunk_b)
    return(info, wave)


def identify(file):
    """wavetype = identify(file)
    
    Identify the type of WAV file and return its type. Files that are
    unable to be identified are classified as generic. The wave type
    identification allows the extraction of proprietary metadata stored
    in the file and filename.
    Input parameter
    
      file - filehandle for the WAV file to be identified
            
    Output

      wavetype - string with the name of the wave file type."""
    import warnings
    from os.path import basename
    import struct
    wavetype = None
    # Test each known file type for type-specific properties.
    # First make sure the file is a WAVE file.
    file.seek(0)
    data = file.read(4)
    if data != b'RIFF':
        if data[0:3] == b'ID3':
            # There is an ID3 tag at the beginning if the file.
            # Only ID3v2 tags are supported. See <https://id3.org/id3v2.3.0>.
            file.seek(2, 1)
            data0 = file.read(1)
            data1 = file.read(1)
            data2 = file.read(1)
            data3 = file.read(1)
            tag_size = (
                int.from_bytes(data3, 'big') + int.from_bytes(data2,
                'big') * 2**7 + int.from_bytes(data1, 'big') * 2**14 +
                int.from_bytes(data0, 'big') * 2**21
           )
            # Skip forward tag_size bytes.
            file.seek(tag_size, 1)
            data = file.read(4)
            if data != b'RIFF':
                raise ValueError('Chunk ID not RIFF: ', data)
        else:
            raise ValueError('Chunk ID not RIFF: ', data)
    file.seek(4, 1)
    data = file.read(4)
    if data != b'WAVE':
        raise ValueError('Chunk ID not WAVE: ', data)
    data = file.read(4)
    # Check to see if the file was written by an icListen device.
    if data == b'LIST':
        # Looks like an icListen file.  Check to see if it has an INFO chunk.
        file.seek(4, 1)
        data = file.read(4)
        if data == b'INFO':
            wavetype = 'icListen'
        else:
            message = 'Not icListen...Chunk Format not INFO...Using generic format'
            warnings.warn(message)
            wavetype = 'generic'
    # Check to see if the this is a bext chunk (used by ZOOM devices).
    elif data == b'bext':
        file.seek(260, 1)
        data = file.read(32)
        tstr = data.split(b'\0',1)[0].split(b' ')[0]
        if tstr == b'ZOOM':
            wavetype = 'zoom'
        else:
            message = 'bext originator string not ZOOM...Using generic format'
            warnings.warn(message)
            wavetype = 'generic'
    # If the 'fmt ' chunk is first, it could be an AudioMoth or a
    # generic WAVE file. Note that Decimus and LARS recordings are
    # generic WAVE files with data encoded in the file name.
    elif data == b'fmt ':
        # AudionMoth files have a LIST chunk with ICMT and IART
        # subchunks just after the fmt  chunk.
        size = struct.unpack('<I', file.read(4))[0]
        file.seek(size, 1)
        data = file.read(4)
        if data == b'LIST':
            # In the LIST chunk. Could be an AudioMoth file.
            size = struct.unpack('<I', file.read(4))[0]
            data = file.read(8)
            if data == b'INFOICMT':
                # Still could be an AudioMoth recording.
                size = struct.unpack('<I', file.read(4))[0]
                icmt = file.read(size).decode('ascii').rstrip('\x00')
                if search(r'AudioMoth', icmt) != None:
                    # This is an AudioMoth recording.
                    wavetype = 'AudioMoth'
        if wavetype == None:
            # Check to see if it is a Decimus file. Decimus files have
            # names beginning with 'PAM'
            if basename(file.name)[0:3] == 'PAM':
                wavetype = 'decimus'
            # Check to see if it is a LARS file. LARS files have names
            # beginning with 'LH'
            elif basename(file.name)[0:2] == 'LH':
                wavetype = 'lars'
            else:
                # Use generic WAVE file format.
                wavetype = 'generic'
    else:
        # File does not match known recorders. Use generic.
        wavetype = 'generic'
    return(wavetype)

def getInfo(file, wavetype=None):
    """Read the file information from the WAV file with handle file and
    return the info dictionary.
    """
    if wavetype == None:
        wavetype = identify(file)
    info = {'wavetype' : wavetype}
    if wavetype == 'AudioMoth':
        from .recorders.AudioMoth import get_info
    if wavetype == 'decimus':
        from .recorders.decimus import get_info
    if wavetype == 'icListen':
        from .recorders.icListen import get_info
    if wavetype == 'lars':
        from .recorders.lars import get_info
    if wavetype == 'zoom':
        from .recorders.zoom import get_info
    if wavetype == 'generic':
        from .recorders.generic import get_info
    info = get_info(file, info)
    return(info)
    
def wave_chunk(file, info, t0=0, t1=-1, chunk_b=3072, verbose=False):
    """Read a WAVE file in chunks (not all at once) and return all the
    data. This is a back-end to the read function.
    """
    import numpy as np
    import struct
    import logging
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(levelname)8s %(name)s | %(message)s')
    ch.setFormatter(formatter)
    logger = logging.getLogger('wave_chunk')
    logger.addHandler(ch)
    if verbose:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)
    t0_pos = info['data0'] + np.uint32(np.floor(t0 * info['fs']) * 
        info['block_align'])
    lastdata_pos = info['data0'] + info['Nsamples'] * info['block_align']
    if t1==-1:
        t1_pos = lastdata_pos
    else:
        # Subtract 1 from t1 address to only include samples up to (not
        # including) t1.
        t1_pos = info['data0'] + np.uint32(np.floor(t1 * info['fs'] - 1) * 
            info['block_align'])
    if (t1_pos > lastdata_pos):
        try:
            raise ValueError('Past EOF')
        except ValueError:
            print("t1 is past the end of file")
            print("t1_pos: ", t1_pos)
            print("lastdata_pos: ", lastdata_pos)
            raise
    # nn is the number of bytes to read.
    nn = t1_pos - t0_pos
    logger.info('wave_chunk reading ' + str(nn) + 'bytes')
    #nnchunk is the number of full chunks to read.
    nnchunk = nn // chunk_b
    logger.info(str(nnchunk) + ' chunks')
    # ncrumb is the leftovers that did not fit in the last chunk.
    ncrumb = nn % chunk_b
    fmt0 = '<' + str(chunk_b * 8 // info['bits'])
    # Determine how to decode data.
    if info['bits'] == 8:
        wave = np.array([], dtype = np.int8)
        decode = lambda bindat: struct.unpack(fmt0 + 'b', bindat)
    elif info['bits'] == 16:
        wave = np.array([], dtype = np.int16)
        decode = lambda bindat: struct.unpack(fmt0 + 'h', bindat)
    elif info['bits'] == 24:
        wave = np.array([], dtype = np.int32)
        decode = _decode24
    elif info['bits'] == 32:
        if info['compress'] == 1:
            # 32 bit signed integer
            wave = np.array([], dtype = np.int32)
            decode = lambda bindat: struct.unpack(fmt0 + 'i', bindat)
        elif info['compress'] == 3:
            # 32 bit float
            wave = np.array([], dtype = np.float32)
            decode = lambda bindat: struct.unpack(fmt0 + 'f', bindat)
        else:
            raise ValueError('Not a recognized format. Bits: ', 
                info['bits'], ' Compress: ', info['compress'])
    elif info['bits'] == 64:
        if info['compress'] == 3:
            # 64 bit float
            wave = np.array([], dtype = np.float64)
            decode = lambda bindat: struct.unpack(fmt0 + 'd', bindat)
        else:
            raise ValueError('Not a recognized format. Bits: ', 
                info['bits'], ' Compress: ', info['compress'])
    else:
        raise ValueError('Not a recognized format. Bits: ', 
            info['bits'], ' Compress: ', info['compress'])
    # Now start reading the data.
    file.seek(t0_pos)
    for nchunk in range(nnchunk):
        logger.info('  Chunk ' + str(nchunk) + ' of ' + str(nnchunk))
        chunk = file.read(chunk_b)
        wave = np.append(wave, decode(chunk))
    if ncrumb:
        # Check to see that ncrumb is a whole number of samples
        if (np.mod(ncrumb, info['bits']//8) > 0):
            # ncrumb has a partial sample at the end likely due to truncation error.
            logger.info('Partial sample in ncrumb:')
            logger.info('  {:d} bytes out of {:d} bytes'.format(np.mod(ncrumb, \
                info['bits']//8), info['bits']//8))
            if (t1 == -1 or t1 >= (info['Nsamples'] - 1) / info['fs']):
                # The last sample in the crumb is the last sample in the file.
                # Reduce ncrumb to use the last complete sample.
                ncrumb = np.int32(np.floor(ncrumb / info['bits']/8)) * \
                    info['bits']//8
                logger.info('Truncating ... ncrumb now {:d}'.format(ncrumb))
            else:
                # There is at least one more sample available in the
                # file. Use it to complete ncrumb.
                ncrumb = np.int32(np.ceil(ncrumb / info['bits']/8)) * \
                    info['bits']//8
                logger.info('Extending ... ncrumb now {:d}'.format(ncrumb))
        logger.info('  Cleaning up ' + str(ncrumb) + ' bytes')
        fmt0 = '<' + str(ncrumb * 8 // info['bits'])
        crumb = file.read(ncrumb)
        wave = np.append(wave, decode(crumb))
    logger.info('wave_chunk done')
    if info['chan'] > 1:
        # Multichannel file. Consecutive samples are in different
        # channels. Reshape with channels in different rows.
        logger.info('Sorting channels')
        wave = np.reshape(wave, (info['chan'], -1), order='F')
    logger.removeHandler(ch)
    return(wave)

def _decode24(string):
    """Decode 24-bit binary integer data, often used in WAV files."""
    from struct import unpack
    fmt = '<'+ str(len(string)//3) + 'i'
    # Convert the string to a 32 bit (4-byte) string by adding \0 as the
    # most significant byte for positive numbers and \xff as the most
    # significant byte for negative numbers.
    str32=b''
    for n in range(len(string))[:-1:3]:
        str32 += string[n:n+3] + (b'\0' if string[n+2] < 128 else b'\xff')
    # Now use the 32-bit unpack function ('i' format) to get the number.
    return(unpack(fmt, str32))



"""
Sgwave is a package that processes and stores spectrogram data in HDF5
format data files. There are functions to process WAV files and store
the spectrograms in the data files and functions to extract spectrograms
from the data files to perform various analyses.

2017-09-26 23:05
"""
from __future__ import division

def sgwavfile(wavfile, t0=0, t1=-1, nfft=2048, noverlap=-1, window='hann', \
    flength=60., ftime=True, datestring=None, ftype=None, verbose=False):
    """Compute spectrograms of a .WAV file and save it to HDF5 files
    (with bzip2 compression implemented). This function uses
    scipy.signal.spectrogram to compute the spectrograms of the data.
    
    sgwavfile(wavfile, t0, t1, nfft, noverlap, window, flength, ftime, ftype, verbose)
    
    Input Parameters
    wavfile: filename of the WAV file to process
    t0: start time (in seconds) for processing. The default value is 0.
    t1: end time (in seconds) for processing. Use -1 for the end of
        the file. The default value is -1.
    nfft: number of points in each FFT. The default value is 2048.
    noverlap: the number of waveform points to overlap in each FFT.
        Use -1 for nfft/2. The default value is -1.
    window: FFT window function or function name as required by
        scipy.signal.spectrogram. The default value is 'hann' (for a
        Hanning window).
    flength: length of each spectrogram file. May either be in time
        (seconds) or samples. The default value is 60 seconds.
    ftime: boolean that specifies whether flength is a time (True) or
        a number of samples (False). The default value is True.
    datestring: set the datestring for the sprctrogram file. If
        datestring is None the datestring is extracted from the WAV
        file. The default value is None.
    ftype: The type of WAV file. This parameter allows the extraction
        of proprietary metadata recorded in some WAV files by
        different recorder. Use None for a standard WAV file,
        'icListen' for an icListen file, and 'zoom' for a Zoom
        recorder WAV file. The default value is False.
    verbose: enable verbose output. The default value is None.
    
    Output
    The output files use the same name as the input files with a
    "_n_sg.h5." with (n=file number) appended before the file extension.
    If the input file is 'fname.wav', and there are 10 spectrogram files
    generated from it, the output files will be named 'fname_0_sg.h5'
    through 'fname_9_sg.h5'. The sgwavfile function returns a value of 0
    upon successful completion.
    """
    import numpy as np
    import scipy.signal as sig
    import tables
    from os.path import exists, basename, dirname
    from os import stat
    from SVSound import wavefile
    import logging
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(levelname)8s %(name)s | %(message)s')
    ch.setFormatter(formatter)
    logger = logging.getLogger('sgwavfile')
    logger.addHandler(ch)
    if verbose:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)
    if (noverlap == -1):
        noverlap = nfft//2
    # Store each sample at a float64.
    atom = tables.Float64Atom()
    # This is the compression algorithm for the output file.
    filters = tables.Filters(complevel=9, complib='bzip2', shuffle=True, \
                fletcher32=True)
    with open(wavfile, 'rb') as infile:
        info = wavefile.getInfo(infile, ftype)
        if datestring == None:
            # Get the datestring from the WAV file.
            if (info['wavetype'] == 'zoom'):
                # Time information is stored in the WAV file in the time
                # zone for which the recorder was set.
                datestring = (info['OriginationDate'] + 'T' +
                    info['OriginationTime'])
            elif (info['wavetype'] == 'icListen'):
                # Extract the time and date information from the
                # filename.
                tdinfo = wavfile.replace('.', '_').split('_')[1:3]
                datestring = (tdinfo[0][:4]+ '-' + tdinfo[0][4:6] +
                    '-' + tdinfo[0][6:8] + 'T' + tdinfo[1][:2] + ':'
                    + tdinfo[1][2:4] + ':' + tdinfo[1][4:6] +
                    'Z')
            elif (info['wavetype'] == 'decimus'):
                # Extract the time and date information from the
                # filename.
                tdinfo = wavfile.replace('.', '_').split('_')[1:4]
                datestring = (tdinfo[0][:4]+ '-' + tdinfo[0][4:6] +
                    '-' + tdinfo[0][6:8] + 'T' + tdinfo[1][:2] + ':'
                    + tdinfo[1][2:4] + ':' + tdinfo[1][4:6] + '.' +
                    tdinfo[2] + 'Z')
            else:
                # Try to get datestring from WAV file modification time.
                fm = stat(wavfile).st_mtime
                datestring = (str( 
                    np.datetime64(np.int64(fm), 's') - 
                        np.timedelta64( 
                            np.int64( 
                                np.round( 
                                    info['Nsamples'] / info['fs'] 
                                ) 
                            ), 
                            's' 
                        ) 
                    ) + 
                    'Z')
        # flength determines the size of the input file slice in seconds
        # (ftime==True) or samples (ftime ==False) for each output file.
        # Note that there will also be some overlap samples at the end
        # of intermediate input files.
        if ftime==False:
            # flength is in samples. Convert to seconds for
            # compatibility with other code.
            flength = flength / info['fs']
        # n0 is the initial sample to process.
        n0 = int(t0 * info['fs'])
        # n1 is the last sample to process.
        if (t1 == -1):
            n1 = info['Nsamples'] - 1
        else:
            n1 = int(t1 * info['fs']) - 1
        # Determine the number of samples to process.
        nn = n1 - n0 + 1
        # Determine the number of output files. There may be extra
        # samples at the end.
        nfiles = np.int64(np.floor(nn / info['fs'] / flength))
        # nnfile is the number of wave samples in each file, not
        # including overlap.
        nnfile = np.int64(flength * info['fs'])
        # nskip is the number of samples to skip for each FFT.
        nskip = nfft - noverlap
        # nnfft is the number of FFTs to process in each complete input
        # file.
        nnfft = np.int64(np.ceil((flength * info['fs'] - nfft) /
            nskip + 1))
        # nend is the number of samples from the first input file sample
        # to the last input file sample to process in each output file.
        # There could be some extra samples past flength to complete the
        # last FFT.
        nend = (nnfft - 1) * nskip + nfft - 1
        wavdir = dirname(wavfile)
        wavbase = basename(wavfile)
        if (wavdir == ''):
            outdir = ''
        else:
            outdir = wavdir + '/'
        for nfile in range(nfiles - 1):
            # Name of output file.
            outfile = outdir + wavbase.rsplit('.', 1)[0] + \
                '_{0:=-02d}_sg.h5'.format(np.int64(t0 / flength) + \
                nfile)
            logger.info('Infile: {}, Outfile: {}'.format(wavfile,
                outfile))
            with tables.open_file(outfile, mode='w', filters=filters) \
                as fout:
                out_t = fout.create_earray(fout.root, 't', atom,
                    (0,), title='Time (s)', expectedrows=nnfft)
                out_f = fout.create_earray(fout.root, 'f', atom,
                    (0,), title='Frequency (Hz)', expectedrows=nnfft)
                out_sg = fout.create_earray(fout.root, 'sg', atom,
                    (nfft//2+1, 0), title='Spectrogram',
                    expectedrows=nnfft)
                out_sg.attrs.inputfile = wavfile
                out_t.attrs.datestring = datestring
                out_sg.attrs.fs = info['fs']
                out_sg.attrs.nfft = nfft
                out_sg.attrs.noverlap = noverlap
                out_sg.attrs.window = window
                # The following is right because nend is the number of
                # input file samples used, not an absolute sample
                # number.
                out_sg.attrs.Nsamples = nend + 1
                t0f = t0 + (nfile * nnfile) / info['fs']
                logger.info('Infile: {}, t0f: {}'.format(wavfile,
                    t0f))
                t1f = t0f + nend / info['fs']
                logger.info('Infile: {}, t1f: '.format(wavfile, t1f))
                wave = wavefile.wave_chunk(infile, info, t0 = t0f, 
                    t1 = t1f, verbose=verbose)
                logger.info(('Infile: {}, Returned from '
                    + 'wave_chunk').format(wavfile)) 
                # Apply the calibration value if it exists.
                if ('cal' in info):              
                    f, t, sg = sig.spectrogram(np.multiply(wave,
                        info['cal']), info['fs'], window='hann',
                        nperseg=nfft, noverlap=noverlap)
                # Else, use the waveform as is in file. 
                else:
                    f, t, sg = sig.spectrogram(wave, info['fs'],
                        window='hann', nperseg=nfft,
                        noverlap=noverlap)                
                out_t.append(t + t0f)
                out_t.flush()
                out_f.append(f)
                out_f.flush()
                out_sg.append(sg)
                out_sg.flush()
        # Take care of last file so it does not pass the end of the
        # input file.
        nfile = nfiles - 1
        outfile = outdir + wavbase.rsplit('.', 1)[0] + \
            '_{0:=-02d}_sg.h5'.format(np.int64(t0 / flength) + \
            nfile)
        logger.info('Infile: {}, Outfile: {}'.format(wavfile,
            outfile))
        with tables.open_file(outfile, mode='w', filters=filters) as \
            fout:
            out_t = fout.create_earray(fout.root, 't', atom, (0,),
                title='Time (s)', expectedrows=nnfft)
            out_f = fout.create_earray(fout.root, 'f', atom, (0,), 
                title='Frequency (Hz)', expectedrows=nnfft)
            out_sg = fout.create_earray(fout.root, 'sg', atom,
                (nfft//2+1, 0), title='Spectrogram',
                expectedrows=nnfft)
            out_sg.attrs.inputfile = wavfile
            out_t.attrs.datestring = datestring
            out_sg.attrs.fs = info['fs']
            out_sg.attrs.nfft = nfft
            out_sg.attrs.noverlap = noverlap
            out_sg.attrs.window = window
            out_sg.attrs.Nsamples = nend + 1
            t0f = t0 + (nfile * nnfile) / info['fs']
            logger.info('Infile: {}, t0f: {}'.format(wavfile, t0f))
            t1f = t0f + nend / info['fs']
            # If the extra samples to complete the last FFT exceed the
            # end of the file, reduce neng by nskip and adjust.
            if (t1f * info['fs'] > info['Nsamples']):
                nend = nend - nskip
                t1f = t0f + nend / info['fs']
            logger.info('Infile: {}, t1f: '.format(wavfile, t1f))
            wave = wavefile.wave_chunk(infile, info, t0=t0f, t1=t1f,
                verbose=verbose)
            logger.info(('Infile: {}, Returned from '
                + 'wave_chunk').format(wavfile)) 
            # Apply the calibration value if it exists.
            if ('cal' in info):              
                f, t, sg = sig.spectrogram(np.multiply(wave,
                    info['cal']), info['fs'], window='hann',
                    nperseg=nfft, noverlap=noverlap)
            # Else, use the waveform as is in file. 
            else:
                f, t, sg = sig.spectrogram(wave, info['fs'],
                    window='hann', nperseg=nfft,
                    noverlap=noverlap)                
            out_t.append(t + t0f)
            out_t.flush()
            out_f.append(f)
            out_f.flush()
            out_sg.append(sg)
            out_sg.flush()
        # Take care of samples at the end.
        nnlast = nn - nfiles * nnfile
        if (nnlast >= nfft):
            # nnfft is the number of FFTs to process in the input file.  
            # Using integer math to truncate.
            nnfftlast = np.int64(np.floor((nnlast - nfft) / nskip + 1))
            # nend is the last input file sample to process.
            nend = (nnfftlast - 1) * nskip + nfft - 1
            outfile = wavfile.rsplit('.', 1)[0] + \
                '_{0:=-02d}_sg.h5'.format(np.int64(t0 / flength) + \
                nfiles)
            logger.info('Infile: {}, Outfile: {}'.format(wavfile,
                outfile))
            with tables.open_file(outfile, mode='w', filters=filters) \
                as fout:
                out_t = fout.create_earray(fout.root, 't', atom,
                    (0,), title='Time (s)', expectedrows=nnfftlast)
                out_f = fout.create_earray(fout.root, 'f', atom,
                    (0,), title='Frequency (Hz)',
                    expectedrows=nnfftlast)
                out_sg = fout.create_earray(fout.root, 'sg', atom,
                    (nfft//2+1, 0), title='Spectrogram',
                    expectedrows=nnfftlast)
                out_t.attrs.datestring = datestring
                out_t.attrs.unit = 's'
                out_f.attrs.unit = 'Hz'
                out_sg.attrs.inputfile = wavfile
                out_sg.attrs.fs = info['fs']
                out_sg.attrs.nfft = nfft
                out_sg.attrs.noverlap = noverlap
                out_sg.attrs.window = window
                out_sg.attrs.Nsamples = nend + 1
                out_sg.attrs.unit = chr(956) + 'Pa^2/Hz'
                out_sg.attrs.composite = False
                t0f = t0 + ((nfiles - 1) * nnfile) / info['fs']
                logger.info('Infile: {}, t0f: {}'.format(wavfile,
                    t0f))
                t1f = t0f + nend / info['fs']
                logger.info('Infile: {}, t1f: {}'.format(wavfile,
                    t1f))
                wave = wavefile.wave_chunk(infile, info, t0=t0f,
                    t1=t1f, verbose=verbose)
#               logger.info('wave read. wave.size: ', wave.size)
                # Apply the calibration value if it exists.
                if ('cal' in info):              
                    f, t, sg = sig.spectrogram(np.multiply(wave,
                        info['cal']), info['fs'], window='hann',
                        nperseg=nfft, noverlap=noverlap)
                # Else, use the waveform as is in file. 
                else:
                    f, t, sg = sig.spectrogram(wave, info['fs'],
                        window='hann', nperseg=nfft,
                        noverlap=noverlap)                
#               logger.info('sg complete.')
#               logger.info('  f.size: ', f.size)
#               logger.info('  t.size: ', t.size)
#               logger.info('  sg.size: ', sg.size)
#               logger.info('')
                out_t.append(t + t0f)
                out_t.flush()
                out_f.append(f)
                out_f.flush()
                out_sg.append(sg)
                out_sg.flush()
            logger.info('Infile: {}, Processing ' + \
                'complete.'.format(wavfile))
            logger.info(('Infile: {}, {:d} input file samples ' + \
                'processed.').format(wavfile, nn))
            logger.info(('Infile: {}, {:d} files ' 
                + 'written').format(wavfile, nfiles + 1))
            logger.info(('Infile: {}, {:d} FFTs in files 0 -'
                + ' {:d}').format(wavfile, nnfft, nfiles - 1))
            logger.info(('Infile: {}, {:d} FFTs in file '
                + '{:d}').format(wavfile, nnfftlast,                    
                nfiles))
        else:
            logger.info(('Infile: {}, Processing '
                + 'complete.').format(wavfile))
            logger.info(('Infile: {}, {:d} input file samples '
                + 'processed.').format(wavfile, nn))
            logger.info(('Infile: {}, {:d} files '
                + 'written').format(wavfile, nfiles))
            logger.info(('Infile: {}, {:d} FFTs in each '
                + 'file').format(wavfile, nnfft))
    # Return a value of 0 for successful completion.
    logger.removeHandler(ch)
    return(0)

def get_sginfo(sgfile):
    import numpy as np
    import tables
    with tables.open_file(sgfile, mode='r') as infile:
        sginfo = {}
        if ('fs' in infile.root.sg.attrs._f_list()):
            sginfo['fs'] = infile.root.sg.attrs.fs
        if ('nfft' in infile.root.sg.attrs._f_list()):
            sginfo['nfft'] = infile.root.sg.attrs.nfft
        if ('fs' in infile.root.sg.attrs._f_list()):
            sginfo['noverlap'] = infile.root.sg.attrs.noverlap
        if ('window' in infile.root.sg.attrs._f_list()):
            sginfo['window'] = infile.root.sg.attrs.window
        if ('datestring' in infile.root.t.attrs._f_list()):
            sginfo['datestring'] = infile.root.t.attrs.datestring
        if ('unit' in infile.root.t.attrs._f_list()):
            sginfo['t unit'] = infile.root.t.attrs.unit
        if ('unit' in infile.root.f.attrs._f_list()):
            sginfo['f unit'] = infile.root.f.attrs.unit
        if ('unit' in infile.root.sg.attrs._f_list()):
            sginfo['sg unit'] = infile.root.sg.attrs.unit
        if ('WAV file' in infile.root.sg.attrs._f_list()):
            sginfo['WAV file'] = infile.root.sg.attrs.inputfile
        if ('WAV nsamp' in infile.root.sg.attrs._f_list()):
            sginfo['WAV nsamp'] = infile.root.sg.attrs.Nsamples
        if ('composite' in infile.root.sg.attrs._f_list()):
            sginfo['composite'] = infile.root.sg.attrs.composite
        else:
            sginfo['composite'] = False
    return(sginfo)

def sgread(sgfile, t0=0, t1=-1, f0=0, f1=-1):
    """Read an HDF5 file written by sgwavefile and extract the spectrogram information \
    from it.
    """
    import numpy as np
    import tables
    with tables.open_file(sgfile, mode='r') as infile:
        t = infile.root.t.read()
        f = infile.root.f.read()
        if (t.size == 0):
            return([], [], [])
        if (t1 == -1):
            t1 = t[-1]
        if (f1 == -1):
            f1 = f[-1]
        tb = np.logical_and(t >= t0, t <= t1)
        fb = np.logical_and(f >= f0, f <= f1)
        tout = t[tb]
        t_ind = np.array(np.where(tb))
        t_ind0 = t_ind.min()
        t_ind1 = t_ind.max()
        fout = f[fb]
        f_ind = np.array(np.where(fb))
        f_ind0 = f_ind.min()
        f_ind1 = f_ind.max()
        sgout = infile.root.sg[t_ind0 : t_ind1 + 1, f_ind0 : f_ind1+1]
    return(fout, tout, sgout)

def sgreadsample(sgfile, nt0=0, nt1=-1, nf0=0, nf1=-1):
    """Read an HDF5 file written by sgwavefile and extract the spectrogram information \
    from it on a given time and frequency sample range.
    """
    import numpy as np
    import tables
    with tables.open_file(sgfile, mode='r') as infile:
        t = infile.root.t.read()
        f = infile.root.f.read()
        if (t.size == 0):
            return(np.array([]), np.array([]), np.array([]))
        if (nt1 == -1):
            nt1 = t.size
        if (nf1 == -1):
            nf1 = f.size
        if nt0 > t.size:
            try:
                raise ValueError('Index nt0 > maximum time sample')
            except Exception as e:
                e.add_note('nt0: {:}'.format(nt0))
                e.add_note('t.size: {:}'.format(t.size))
                raise
        if nt1 > t.size:
            try:
                raise ValueError('Index nt1 > maximum time sample')
            except Exception as e:
                e.add_note('nt1: {:}'.format(nt1))
                e.add_note('t.size: {:}'.format(t.size))
                raise
        if nf0 > f.size:
            try:
                raise ValueError('Index nf0 > maximum frequency sample')
            except Exception as e:
                e.add_note('nf0: {:}'.format(nf0))
                e.add_note('f.size: {:}'.format(f.size))
                raise
        if nf1 > f.size:
            try:
                raise ValueError('Index nf1 > maximum frequency sample')
            except Exception as e:
                e.add_note('nf1: {:}'.format(nf1))
                e.add_note('f.size: {:}'.format(f.size))
                raise
        fout = f[nf0:nf1]
        tout = t[nt0:nt1]      
        sgout = infile.root.sg[nt0:nt1, nf0:nf1]
    return(fout, tout, sgout)

    
def compsg(sgfiles, t0=0, t1=-1, avtime=-1, savefile=False, foutname=None):
    """Produce a composite spectrogram from data files written by sgwavfile.
    """
    import numpy as np
    from os.path import exists, basename, dirname
    if isinstance(sgfiles, str):
        # There is only one file not in a list. Put it in a list.
        sgfiles = [sgfiles]
    elif not (isinstance(sgfiles, list) or isinstance(sgfiles, tuple)):
        raise ValueError('Parameter not list, tuple, or string', sgfiles)
    taps = np.array([], dtype=np.float64)
    datet = np.array([], dtype='M8[ns]')
    # Process first file.
    info = get_sginfo(sgfiles[0])
    if (type(info['datestring']) == np.bytes_) or \
        (type(info['datestring']) == bytes):
        info['datestring'] = info['datestring'].decode('utf-8')
    # Note: We are assuming that the utc value for all files is the same.
    if (info['datestring'][-1] == 'Z'):
        # datestring is in UTC. Avoid depreciation warning in np.datetime64.
        utc = True
        datet0 = np.datetime64(info['datestring'][:-1])
    else:
        # datestring is in local time.
        utc = False
        datet0 = np.datetime64(info['datestring'])
    nskip = info['nfft'] - info['noverlap']
    dt = nskip / info['fs']
    if (len(sgfiles) == 1):
        f, t, sg = sgread(sgfiles[0], t0=t0, t1=t1)
    else:
        f, t, sg = sgread(sgfiles[0], t0=t0, t1=-1)
    t0f = t[0] - dt
    # Determine the number of average power spectra in for the file.
    if (avtime == -1):
        taps = np.append(taps, (t[0] + t[-1]) / 2)
        datet = np.append(datet, datet0 + \
            np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
        psav = np.sum(sg, axis=1).reshape(f.size, 1) / t.size
        csg = psav
    else:
        # Find the number of spectrograms for which the last spectrogram is based on
        # samples entirely within the avtime window.
        avnn = np.int64(np.floor((avtime * info['fs'] - info['nfft']) / 
            (info['nfft'] - info['noverlap']) + 1))
        # nnpsav is the number of average power spectra in the file.
        nnpsav = np.int64(np.floor(t.size / avnn))
        for n in range(nnpsav):
            nst0 = n * avnn
            nst1 = (n + 1) * avnn
            #if (nst1 >= t.size):
            #    nst1 = t.size
            taps = np.append(taps, (t[nst0] + t[nst1]) / 2.0)
            datet = np.append(datet, datet0 + \
                np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
            psav = np.sum(sg[:, nst0:nst1], axis=1).reshape(f.size, 1) / (nst1 - nst0)
            if (n == 0):
                csg = psav
            else:
                csg = np.append(csg, psav, axis=1)
    # Process the intermediate files.
    for sgfile in sgfiles[1:-1]:
        info = get_sginfo(sgfile)
        if (type(info['datestring']) == np.bytes_) or \
            (type(info['datestring']) == bytes):
            info['datestring'] = info['datestring'].decode('utf-8')

        if (info['datestring'][-1] == 'Z'):
            # datestring is in UTC. Avoid depreciation warning in np.datetime64.
            datet0 = np.datetime64(info['datestring'][:-1])
        else:
            # datestring is in local time.
            datet0 = np.datetime64(info['datestring'])
        nskip = info['nfft'] - info['noverlap']
        dt = nskip / info['fs']
        f, t, sg = sgread(sgfile, t0=0, t1=-1)
        t0f = t[0] - dt
        # Determine the number of average power spectra in for the file.
        if (avtime == -1):
            taps = np.append(taps, (t[0] + t[-1]) / 2.0)
            datet = np.append(datet, datet0 + \
                np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
            psav = np.sum(sg, axis=1).reshape(f.size, 1) / t.size
            csg = np.append(csg, psav, axis=1)
        else:
            # Find the number of spectrograms for which the last spectrogram is based on
            # samples entirely within the avtime window.
            avnn = np.int64(np.floor((avtime * info['fs'] - info['nfft']) / 
                (info['nfft'] - info['noverlap']) + 1))
            nnpsav = np.int64(np.floor(t.size / avnn))
            for n in range(nnpsav):
                nst0 = n * avnn
                nst1 = (n + 1) * avnn
                #if (nst1 >= t.size):
                #    nst1 = t.size
                taps = np.append(taps, (t[nst0] + t[nst1]) / 2.0)
                datet = np.append(datet, datet0 + \
                    np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
                psav = np.sum(sg[:, nst0:nst1], axis=1).reshape(f.size, 1) / \
                    (nst1 - nst0)
                csg = np.append(csg, psav, axis=1)
    if (len(sgfiles) > 1):
        # Process the last file.
        info = get_sginfo(sgfiles[-1])
        if (type(info['datestring']) == np.bytes_) or \
            (type(info['datestring']) == bytes):
            info['datestring'] = info['datestring'].decode('utf-8')
        if (info['datestring'][-1] == 'Z'):
            # datestring is in UTC. Avoid depreciation warning in np.datetime64.
            datet0 = np.datetime64(info['datestring'][:-1])
        else:
            # datestring is in local time.
            datet0 = np.datetime64(info['datestring'])
        nskip = info['nfft'] - info['noverlap']
        dt = nskip / info['fs']
        f, t, sg = sgread(sgfiles[-1], t0=0, t1=t1)
        t0f = t[0] - dt
        # Determine the number of average power spectra in for the file.
        if (avtime == -1):
            taps = np.append(taps, (t[0] + t[-1]) / 2.0)
            datet = np.append(datet, datet0 + \
                np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
            psav = np.sum(sg, axis=1).reshape(f.size, 1) / t.size
            csg = np.append(csg, psav, axis=1)
        else:
            # Find the number of spectrograms for which the last spectrogram is based on
            # samples entirely within the avtime window.
            avnn = np.int64(np.floor((avtime * info['fs'] - info['nfft']) / 
                (info['nfft'] - info['noverlap']) + 1))
            # nnpsav is the number of average power spectra in the file.
            nnpsav = np.int64(np.floor(t.size / avnn))
            for n in range(nnpsav):
                nst0 = n * avnn
                nst1 = (n + 1) * avnn
                #if (nst1 >= t.size):
                #    nst1 = t.size
                datet = np.append(datet, datet0 + \
                    np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
                taps = np.append(taps, (t[nst0] + t[nst1]) / 2.0)
                psav = np.sum(sg[:, nst0:nst1], axis=1).reshape(f.size, 1) / \
                    (nst1 - nst0)
                csg = np.append(csg, psav, axis=1)
    if savefile:
        if (not foutname):
            foutname = basename(sgfiles[0]).rsplit('_sg', 1)[0] + \
                '_csg.h5'
        compsg_save(foutname, csg, f, taps, datet, utc, info)
    return(f, taps, datet, utc, csg)

def compsg_save(fname, csg, f, t, datet, utc=True, sginfo=False):
    import numpy as np
    import tables
    filters = tables.Filters(complevel=9, complib='bzip2', shuffle=True, \
                fletcher32=True)
    atom = tables.Float64Atom()
    with tables.open_file(fname, mode='w', filters=filters) as fout:
        out_t = fout.create_earray(fout.root, 't', atom, (0,), \
            title='Time (s)', expectedrows=t.size)
        out_datet = fout.create_earray(fout.root, 'datet', atom, (0,), \
            title='Time (s)', expectedrows=datet.size)
        out_f = fout.create_earray(fout.root, 'f', atom, (0,), \
            title='Frequency (Hz)', expectedrows=f.size)        
        out_sg = fout.create_earray(fout.root, 'sg', atom, (csg.shape[0], 0), \
            title='Spectrogram', expectedrows=csg.shape[1])
        if (sginfo):
            sgkeys = sginfo.keys()            
            if 'fs' in sgkeys:
                out_sg.attrs.fs = sginfo['fs']
            if 'nfft' in sgkeys:
                 out_sg.attrs.nfft = sginfo['nfft']
            if 'noverlap' in sgkeys:
                out_sg.attrs.noverlap = sginfo['noverlap']
            if 'window' in sgkeys:
                out_sg.attrs.window = sginfo['window']
            if 't unit' in sgkeys:
                out_t.attrs.unit = sginfo['t unit']
            if 'f unit' in sgkeys:
                out_f.attrs.unit = sginfo['f unit']
            if 'sg unit' in sgkeys:
                out_sg.attrs.unit = sginfo['sg unit']
        out_t.attrs.datestring = str(datet[0])
        out_sg.attrs.composite = True
        out_t.append(t)
        out_t.flush()
        out_datet.append(datet)
        out_datet.flush()
        out_f.append(f)
        out_f.flush()
        out_sg.append(csg)
        out_sg.flush()
        
def compsg_plot_files(sgfiles, t0=0, t1=-1, avtime=-1, flim=False, dtformat=None,
    savefile=False, foutname=None):
    """Use compsg to produce a composite spectrogram and plot it.
    """
    import numpy as np
    import matplotlib.pyplot as plt
    import matplotlib.dates as mdates
    from datetime import datetime
    from os.path import basename, dirname
    import warnings
    f, _, datet, utc, csg = compsg(sgfiles, t0, t1, avtime)
    # Need to convert from np.datetime64 to datetime.datetime to plot
    # with matplotlib.
    dt0 = _dt_convert(datet[0])
    dt1 = _dt_convert(datet[-1])
    x_lims = mdates.date2num([dt0, dt1])
    fig, ax = plt.subplots()
    sg = plt.imshow(10*np.log10(csg), origin='lower', aspect='auto',
        extent=[x_lims[0], x_lims[-1], f[0], f[-1]])
    ax.xaxis_date()
    if (dtformat):
        date_format = mdates.DateFormatter(dtformat)
    else:
        date_format = mdates.DateFormatter('%H:%M')
    ax.xaxis.set_major_formatter(date_format)
    if (utc):
        ax.set_xlabel('UTC Time (HH:MM)')
    else:
        ax.set_xlabel('Time (HH:MM)')
    cb = plt.colorbar(shrink=0.5)
    cb.set_label('Power Spectral Density (dB)')
    if (flim):
      ax.set_ylim(flim)
    fig.tight_layout()
    if (savefile):
        if (not foutname):
            foutname = basename(sgfiles[0]).rsplit('_sg', 1)[0] + \
                '_csg.pdf'
        fig.savefig(foutname)
        plt.close(fig)
    else:
        fig.show()
        return(fig, ax, sg, cb)
    

def psbsum(sgfiles, f0=0, f1=-1, t0=0, t1=-1, avtime=-1,
    savefile=False, foutname=None):
    """Produce a power spectral band sum from data files written by
    sgwavfile.
    """
    import numpy as np
    if isinstance(sgfiles, str):
        # There is only one file not in a list. Put it in a list.
        sgfiles = [sgfiles]
    elif not (isinstance(sgfiles, list) or isinstance(sgfiles, tuple)):
        raise ValueError('Parameter not list, tuple, or string', sgfiles)
    taps = np.array([], dtype=np.float64)
    datet = np.array([], dtype='M8[ns]')
    psbsum = np.array([], dtype=np.float64)
    # Process first file.
    info = get_sginfo(sgfiles[0])
    if (type(info['datestring']) == np.bytes_) or \
        (type(info['datestring']) == bytes):
        info['datestring'] = info['datestring'].decode('utf-8')
    # Note: We are assuming that the utc value for all files is the same.
    if (info['datestring'][-1] == 'Z'):
        # datestring is in UTC. Avoid depreciation warning in np.datetime64.
        utc = True
        datet0 = np.datetime64(info['datestring'][:-1])
    else:
        # datestring is in local time.
        utc = False
        datet0 = np.datetime64(info['datestring'])
    nskip = info['nfft'] - info['noverlap']
    dt = nskip / info['fs']
    if (len(sgfiles) == 1):
        f, t, sg = sgread(sgfiles[0], t0=t0, t1=t1)
    else:
        f, t, sg = sgread(sgfiles[0], t0=t0, t1=-1)
    t0f = t[0] - dt
    # nf0 is the index of the lowest frequency band to be included in the sum.
    nf0 = np.int64(np.floor(f0 * info['nfft'] / info['fs']))
    # nf1 is the index of the highest frequency band to be included in the sum.
    if (f1==-1):
        f1 = f[-1]
    nf1 = np.int64(np.floor(f1 * info['nfft'] / info['fs']))
    # Determine the number of average power spectra in for the file.
    if (avtime == -1):
        taps = np.append(taps, (t[0] + t[-1]) / 2.0)
        datet = np.append(datet, datet0 + \
            np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
        psbsum = np.append(psbsum, np.sum(sg[nf0:nf1+1].flatten()) / t.size)
    else:
        # Find the number of spectrograms for which the last spectrogram is based on
        # samples entirely within the avtime window.
        avnn = np.int64(np.floor((avtime * info['fs'] - info['nfft']) / 
            nskip + 1))
        # nnpsav is the number of average power spectra in the file.
        nnpsav = t.size // avnn
        for n in range(nnpsav):
            nst0 = n * avnn
            nst1 = (n + 1) * avnn
            if (nst1 >= t.size):
                nst1 = t.size
            taps = np.append(taps, (t[nst0] + t[nst1-1]) / 2.0)
            datet = np.append(datet, datet0 + \
                np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
            psbsum = np.append(psbsum, np.sum(sg[nf0:nf1+1, \
                nst0:nst1].flatten()) / (nst1 - nst0))
    # Process the intermediate files.
    for sgfile in sgfiles[1:-1]:
        info = get_sginfo(sgfile)
        if (type(info['datestring']) == np.bytes_) or \
            (type(info['datestring']) == bytes):
            info['datestring'] = info['datestring'].decode('utf-8')
        if (info['datestring'][-1] == 'Z'):
            if (not utc):
                warnings.warn('Inconsitent UTC time encoding') 
            # datestring is in UTC. Avoid depreciation warning in np.datetime64.
            datet0 = np.datetime64(info['datestring'][:-1])
        else:
            if (utc):
                warnings.warn('Inconsitent UTC time encoding') 
            # datestring is in local time.
            datet0 = np.datetime64(info['datestring'])
        nskip = info['nfft'] - info['noverlap']
        dt = nskip / info['fs']
        f, t, sg = sgread(sgfile, t0=0, t1=-1)
        t0f = t[0] - dt
        # nf0 is the index of the lowest frequency band to be included in the sum.
        nf0 = np.int64(np.floor(f0 * info['nfft'] / info['fs']))
        # nf1 is the index of the highest frequency band to be included in the sum.
        nf1 = np.int64(np.floor(f1 * info['nfft'] / info['fs']))
        # Determine the number of average power spectra in for the file.
        if (avtime == -1):
            taps = np.append(taps, (t[0] + t[-1]) / 2.0)
            datet = np.append(datet, datet0 + \
                np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
            psbsum = np.append(psbsum, \
                np.sum(sg[nf0:nf1+1].flatten()) / t.size)
        else:
            # Find the number of spectrograms for which the last spectrogram is based on
            # samples entirely within the avtime window.
            avnn = np.int64(np.floor((avtime * info['fs'] - info['nfft']) / 
                nskip + 1))
            # nnpsav is the number of average power spectra in the file.
            nnpsav = t.size // avnn
            for n in range(nnpsav):
                nst0 = n * avnn
                nst1 = (n + 1) * avnn
                if (nst1 >= t.size):
                    nst1 = t.size
                taps = np.append(taps, (t[nst0] + t[nst1-1]) / 2.0)
                datet = np.append(datet, datet0 + \
                    np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
                psbsum = np.append(psbsum, np.sum(sg[nf0:nf1+1, \
                    nst0:nst1].flatten()) / (nst1 - nst0))
    if (len(sgfiles) > 1):
        # Process the last file.
        info = get_sginfo(sgfiles[-1])
        if (type(info['datestring']) == np.bytes_) or \
            (type(info['datestring']) == bytes):
            info['datestring'] = info['datestring'].decode('utf-8')
        if (info['datestring'][-1] == 'Z'):
            if (not utc):
                warnings.warn('Inconsitent UTC time encoding') 
            # datestring is in UTC. Avoid depreciation warning in np.datetime64.
            datet0 = np.datetime64(info['datestring'][:-1])
        else:
            if (utc):
                warnings.warn('Inconsitent UTC time encoding') 
            # datestring is in local time.
            datet0 = np.datetime64(info['datestring'])
        nskip = info['nfft'] - info['noverlap']
        dt = nskip / info['fs']
        f, t, sg = sgread(sgfiles[-1], t0=0, t1=t1)
        t0f = t[0] - dt
        # nf0 is the index of the lowest frequency band to be included in the sum.
        nf0 = np.int64(np.floor(f0 * info['nfft'] / info['fs']))
        # nf1 is the index of the highest frequency band to be included in the sum.
        nf1 = np.int64(np.floor(f1 * info['nfft'] / info['fs']))
        # Determine the number of average power spectra in for the file.
        if (avtime == -1):
            taps = np.append(taps, (t[0] + t[-1]) / 2.0)
            datet = np.append(datet, datet0 + \
                np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
            psbsum = np.append(psbsum, \
                np.sum(sg[nf0:nf1+1].flatten()) / t.size)
        else:
            # Find the number of spectrograms for which the last spectrogram is based on
            # samples entirely within the avtime window.
            avnn = np.int64(np.floor((avtime * info['fs'] - info['nfft']) / 
                (info['nfft'] - info['noverlap']) + 1))
            # nnpsav is the number of average power spectra in the file.
            nnpsav = t.size // avnn
            for n in range(nnpsav):
                nst0 = n * avnn
                nst1 = (n + 1) * avnn
                if (nst1 >= t.size):
                    nst1 = t.size
                taps = np.append(taps, (t[nst0] + t[nst1-1]) / 2.0)
                datet = np.append(datet, datet0 + \
                    np.timedelta64(np.int64(taps[-1] * 1.e6), 'us'))
                psbsum = np.append(psbsum, np.sum(sg[nf0:nf1+1, \
                    nst0:nst1].flatten()) / (nst1 - nst0))
    if savefile:
        if (foutname==-1):
            from os.path import basename
            foutname = basename(sgfiles[0]).rsplit('_sg', 1)[0] + \
                'psb_' + str(int(f0)) + '_' + str(int(f1)) + '.csv'
        psbsum_save(foutname, psbsum, f0, f1, taps, datet, utc, info)
    return(taps, datet, utc, psbsum)

def psbsum_save(fname, psbsum, f0, f1, t, datet, utc=True, sginfo=False, dB=True):
    import numpy as np
    # Make an array of datestrings with ms precision.
    datestr = np.array(np.array(datet, dtype='datetime64[ms]'), dtype=str)
    # Make an output array with rows for each row in the CSV file.
    output = np.zeros((psbsum.size,), dtype=[('datestring', '|S42'), \
        ('t', 'float64'), ('psbsum', 'float64')])
    output['datestring'] = datestr
    output['t'] = t
    if dB:
        output['psbsum'] = 10. * np.log10(psbsum)
    else:
        output['psbsum'] = psbsum   
    header = 'f0: ' + str(f0) + ', f1: ' + str(f1) + '\n' 
    if utc:
        header = header + 'Date String (UTC), Time (s)'
    else:
        header = header + 'Date String (local time), Time (s)'        
    if (dB):
        header = header + ', PSB sum (dB)'
    else:
        header = header + ', PSB sum'
    np.savetxt(fname, output, fmt='%s, %.18g, %.18g', delimiter=',', header=header)

def psbsum_plot_files(sgfiles, f0=0, f1=-1, t0=0, t1=-1, avtime=-1, dB=True,
    dtformat=None, savefile=False, foutname=None):
    """Use psbsum to produce a power spectral band sum and plot it.
    """
    import numpy as np
    import matplotlib.pyplot as plt
    import matplotlib.dates as mdates
    from datetime import datetime
    from os.path import basename, dirname
    import warnings
    taps, datet, utc, psb = psbsum(sgfiles, f0, f1, t0, t1, avtime)
    # Need to convert from np.datetime64 to datetime.datetime to plot
    # with matplotlib.
    dtdatet = np.array([])
    for adatet in datet:
        dtdatet = np.append(dtdatet, _dt_convert(adatet))
    fig, ax = plt.subplots()
    if (dB):
        psbplot = plt.plot(dtdatet, 10*np.log10(psb), '-k')
        ax.set_ylabel('Power Spectral Density (dB)')
    else:
        psbplot = plt.plot(dtdatet, psb, '-k')
        ax.set_ylabel('Power Spectral Density')
    ax.xaxis_date()
    if (dtformat):
        date_format = mdates.DateFormatter(dtformat)
    else:
        if ((datet[-1] - datet[0]).astype('m8[D]') >
            np.timedelta64(0,'D')):
            # There is more than one day of data. Include date.
            date_format = mdates.DateFormatter('%Y-%m-%d %H:%M')
        else:
            date_format = mdates.DateFormatter('%H:%M')
    ax.xaxis.set_major_formatter(date_format)
    if (utc):
        ax.set_xlabel('UTC Time')
    else:
        ax.set_xlabel('Time')
    if (f1 == -1):
        ax.set_title('Frequency: > {0:.6g} Hz'.format(f0))
    else:
        ax.set_title('Frequency: {0:.6g} Hz - {1:.6g} Hz'.format(f0,
            f1)) 
    fig.autofmt_xdate()
    if (savefile):
        if (not foutname):
            foutname = basename(sgfiles[0]).rsplit('_sg', 1)[0] + \
                'psb_' + str(int(f0)) + '_' + str(int(f1)) + '.pdf'
        fig.savefig(foutname)
        plt.close(fig)
    else:
        fig.show()
        return(fig, ax, psbplot)

def multi_psbsum(sgfiles, fbands=[[0, -1]], t0=0, t1=-1, avtime=-1,
    total=True, dB=False, savefile=True, foutname=None):
    """Use psbsum to compute multiple power spectral band sums from the
    same input files.
    """
    import numpy as np
    from datetime import datetime
    from os.path import basename, dirname
    if isinstance(sgfiles, str):
        # There is only one file not in a list. Put it in a list.
        sgfiles = [sgfiles]
    elif not (isinstance(sgfiles, list) or isinstance(sgfiles,
        tuple)):
        raise ValueError('Parameter not list, tuple, or string',
            sgfiles)
    fbnp = np.array(fbands)
    if (fbnp.size == 2):
        # There is only one band entered as a 1-D list.
        fbnp = np.array(fbnp).reshape((1,2))
    if (fbnp.shape[1] != 2):
        raise ValueError('Parameter not a list of 2-value lists',
            fbands)
    if (total):
        fbnp = np.append(fbnp, np.array([[0, -1]]), axis=0)
    nnfb = fbnp.shape[0]
    taps, datet, utc, psb = psbsum(sgfiles, fbnp[0,0], fbnp[0,1], t0,
        t1, avtime)
    multipsb = np.array([psb])
    for nfb in range(1, nnfb):
        _, _, _, psb = psbsum(sgfiles, fbnp[nfb,0], fbnp[nfb,1], t0,
            t1, avtime)
        multipsb = np.append(multipsb, np.array([psb]), axis=0)
    if (savefile):
        if (not foutname):
            if (dirname(sgfiles[0]) == ''):
                foutname = basename(sgfiles[0]).rsplit('_sg', 1)[0] + \
                    'psbsums.csv'
            else:
                foutname = dirname(sgfiles[0]) + '/' + \
                    basename(sgfiles[0]).rsplit('_sg', 1)[0] + \
                    'psbsums.csv'
        # Make an array of datestrings with ms precision.
        datestr = np.array(np.array(datet, dtype='datetime64[ms]'),
            dtype=str)
        # Make an output array with rows for each row in the CSV file.
        dtp = [('datestring', '|S42'), ('t', 'float64')]
        for nfb in range(nnfb):
            dtp = dtp + [('psbsum[{}]'.format(nfb),'<f8')]
        dtp=np.dtype(dtp)
        output = np.zeros((psb.size,), dtype=dtp)
        output['datestring'] = datestr
        output['t'] = taps
        fmt = '%s, %.18g'
        if (dB):
            multipsb = 10. * np.log10(multipsb)
            psbunit= ' (dB)'
        else:
            psbunit = ''
        if utc:
            header = 'Date String (UTC), Time (s)'
        else:
            header = 'Date String (local time), Time (s)'        
        for nfb in range(nnfb):
            if (fbnp[nfb,1] == -1):
                if (fbnp[nfb,0] == 0):
                    header = header + ', Full Spectrum PSB Sum' + \
                        psbunit
                else:
                    header + (', PSB Sum f > {0:.6g} Hz' + \
                psbunit).format(fbnp[nfb,0])
            else:
                header = header + \
                    (', PSB Sum {0:.6g} Hz - {1:.6g} Hz' + \
                    psbunit).format(fbnp[nfb,0], fbnp[nfb,1])
            fmt = fmt + ', %.18g'
            psbkey = 'psbsum[{}]'.format(nfb)
            output[psbkey] = multipsb[nfb]
        np.savetxt(foutname, output, fmt=fmt, delimiter=',',
            header=header)
    else:
        return(taps, datet, utc, multipsb)
        


def _dt_convert(dt64):
    """Convert np.datetime64 ns (dtype('<M8[ns]') object to
    datetime.datettime object. 
    """
    from datetime import datetime
    ns = 1.e-9
    return(datetime.utcfromtimestamp(dt64.astype(int) * ns))



"""Module containing get_info program for a Broadcast Wave file written
by an icListen device.
"""
from __future__ import division

def get_info(file, info={}):
    """info = get_info(file, info={})
    Read the information in a WAV file written by an icListen recorder
    and return the contents. The standard information in the fmt chunk
    is included in the info dictionary along with other information
    encoded in INFO chunk.
    Input Parameters
    
        file - filehandle of an open WAV file
        info - (optional) dictionary that may contain file
            information from other sources. Defaults to an empty
            dictionary.
    
    Output
    
        info - dictionary with information read from the file. If an
            info dictionary was supplied as an input parameter,
            entires that were not changed are also included.
    
    Standard info dictionary keys and values returned
    
        "bits" - integer with the number of bits in each sample.
        "block_align" - number of bytes sampled at the same time (all
            channels combined) in the data
        "byte_per_s" - integer number of bytes per second recorded
        "chan" - integer number of channels in the file
        "compress" - integer Wave file compression index. Only 1
            (uncompressed integer data) and 3 (uncompressed floating
            point data) are currently supported.
        "data0" - integer byte address of the first sample in the file
        "filesize" - integer size of the file in bytes
        "fs" - integer sample rate in samples/second
        "Nsamples" - integer number of samples in the file (in each channel)
        "wavetype" - string with "icListen" as the wave file type read.
        
    Additional info dictionary keys and values returned
    
        Each key/value pair written to the INFO chunk in the file are
        added to the info dictionary. See the icListen documentation for
        details on these parameters. The value info["cal"] contains a
        float64 calibration value for the data. Multiply data samples by
        this value to obtain calibrated values in micropascals.
    """
    import struct
    import numpy as np
    file.seek(0)
    data = file.read(4)
    if data != b'RIFF':
        raise ValueError('Chunk ID not RIFF: ', data)
    data = file.read(4)
    info['filesize'] = struct.unpack('<I',data)[0] + 8
    data = file.read(4)
    if data != b'WAVE':
        raise ValueError('Chunk ID not WAVE: ', data)
    data = file.read(4)
    if data != b'LIST':
        raise ValueError('Chunk ID not LIST: ', data)
    data = file.read(4)
    info['LISTsize'] = struct.unpack('<I',data)[0]
    data = file.read(4)
    if data != b'INFO':
        raise ValueError('Chunk Format not INFO: ', data)
    # Now start reading info chunks and storing them in info{}.
    key = file.read(4)
    while key != b'fmt ':
        akey = key.decode('ascii')
        size = struct.unpack('<I', file.read(4))[0]
        info[akey] = file.read(size).decode('ascii')
        key = file.read(4)
    # We are now in the fmt sub chunk. Set some INFO parameters before 
    # continuing.
    icmt = info['ICMT'].split()
    # Not all icListen recorders use the same order of parameters. 
    # Loop through icmt to get parameters.
    for i in range(len(icmt)):
        if (icmt[i] == 'V') and (icmt[i+1] == 'pk,'):
            vmax = np.float64(icmt[i-1])
        # The following values are encoded as integers but can be forced into 
        # float64 (double precision) with no loss of precision.
        if (icmt[i] == 'dBV') and (icmt[i+1] == 're') and (icmt[i+2]
            == '1uPa,'):
            sens = np.float64(icmt[i-1])
        if (icmt[i] == '=') and (icmt[i+1] == 'Max') and (icmt[i+2]
            == 'Count,'):
            wavemax = np.float64(icmt[i-1])
    info['cal'] = vmax / wavemax * 10**(-sens / 20.0)
    # Continue with fmt sub chunk.
    fmtsize = struct.unpack('<I', file.read(4))[0]
    info['compress'] = struct.unpack('<H', file.read(2))[0]
    info['chan'] = struct.unpack('<H', file.read(2))[0]
    info['fs'] = struct.unpack('<I', file.read(4))[0]
    info['byte_per_s'] = struct.unpack('<I', file.read(4))[0]
    info['block_align'] = struct.unpack('<H', file.read(2))[0]
    info['bits'] = struct.unpack('<H', file.read(2))[0]
    if fmtsize > 16:
        size = struct.unpack('<H', file.read(2))[0]
        info['extra fmt'] = file.read(size)
    data = file.read(4)
    if data != b'data':
        raise ValueError('Chunk ID not data: ', data)
    # We are at the data chunk.
    # Nsamples is the number samples in the file (should be total time * fs).
    info['Nsamples'] = struct.unpack('<I', file.read(4))[0] // \
        (info['block_align'])
    # data0 is the position (in bytes) of the first data sample.
    info['data0'] = file.tell()
    info['wavetype'] = 'icListen'
    return(info)



"""Module containing get_info program for a Broadcast Wave file written
by an AudioMoth device.
"""
from __future__ import division

def get_info(file, info={}):
    """info = get_info(file, info={})
    Read the information in a WAV file written by an AudioMoth recorder
    and return the contents. The standard information in the fmt chunk
    is included in the info dictionary along with other information
    encoded in INFO chunk.
    Input Parameters
    
        file - filehandle of an open WAV file
        info - (optional) dictionary that may contain file
            information from other sources. Defaults to an empty
            dictionary.
    
    Output
    
        info - dictionary with information read from the file. If an
            info dictionary was supplied as an input parameter,
            entires that were not changed are also included.
    
    Standard info dictionary keys and values returned
    
        "bits" - integer with the number of bits in each sample.
        "block_align" - number of bytes sampled at the same time (all
            channels combined) in the data
        "byte_per_s" - integer number of bytes per second recorded
        "chan" - integer number of channels in the file
        "compress" - integer Wave file compression index. Only 1
            (uncompressed integer data) and 3 (uncompressed floating
            point data) are currently supported.
        "data0" - integer byte address of the first sample in the file
        "filesize" - integer size of the file in bytes
        "fs" - integer sample rate in samples/second
        "Nsamples" - integer number of samples in the file (in each channel)
        "wavetype" - string with "AudioMoth" as the wave file type read.
        
    Additional info dictionary keys and values returned
    
        "ICMT" - string with the contents of the ICMT subchunk.
        "IART" - string with the contents of the IART subchunk.
        "datestring" - string with the date and time of the beginning
            of the recording in ISO 8601 format.
        "voltage" - string with the battery voltage at the beginning
            of the recording.
        "gain" - string with the AudioMoth gain setting for
            recording.
        "serial number" - string with the serial number of the
            AudioMoth recording device.      
    """
    
    import struct
    import numpy as np
    import datetime as dt
    import re
    file.seek(0)
    data = file.read(4)
    if data != b'RIFF':
        raise ValueError('Chunk ID not RIFF: ', data)
    data = file.read(4)
    info['filesize'] = struct.unpack('<I',data)[0] + 8
    data = file.read(4)
    if data != b'WAVE':
        raise ValueError('Chunk ID not WAVE: ', data)
    data = file.read(4)
    if data != b'fmt ':
        raise ValueError('Chunk ID not fmt : ', data)
    fmtsize = struct.unpack('<I', file.read(4))[0]
    info['compress'] = struct.unpack('<H', file.read(2))[0]
    info['chan'] = struct.unpack('<H', file.read(2))[0]
    info['fs'] = struct.unpack('<I', file.read(4))[0]
    info['byte_per_s'] = struct.unpack('<I', file.read(4))[0]
    info['block_align'] = struct.unpack('<H', file.read(2))[0]
    info['bits'] = struct.unpack('<H', file.read(2))[0]
    if fmtsize > 16:
        size = struct.unpack('<H', file.read(2))[0]
        info['extra fmt'] = file.read(size)
    data = file.read(4)    
    if data != b'LIST':
        raise ValueError('Chunk ID not LIST: ', data)
    data = file.read(4)
    info['LISTsize'] = struct.unpack('<I',data)[0]
    data = file.read(4)
    if data != b'INFO':
        raise ValueError('Chunk Format not INFO: ', data)
    # The LIST chunk consists of INFO containing ICMT and IART
    # subchunks.
    data = file.read(4)
    while data != b'data':
        if data == b'ICMT':
            # This is the information subchunk. Store in a string.
            size = struct.unpack('<I', file.read(4))[0]
            info['ICMT'] = (
                file.read(size).decode('ascii').rstrip('\x00'))
            # Extract the recording time.
            pattern = (r'(\d\d):(\d\d):(\d\d) ' +
                r'(\d\d)/(\d\d)/(\d\d\d\d) ' +
                r'\(UTC([+-])(\d\d?)\)')
            tmatch = re.search(pattern, info['ICMT'])
            if len(tmatch[8]) == 1:
                tz = tmatch[7] + '0' + tmatch[8] + '00'
            else:
                tz = tmatch[7] + tmatch[8] + '00'
            tstring = (tmatch[6] + tmatch[5] + tmatch[4] + 'T' +
                tmatch[1] + tmatch[2] + tmatch[3] + tz)
            info['datestring'] = dt.datetime.strptime(tstring,
                '%Y%m%dT%H%M%S%z').isoformat()
            # Extract the battery voltage.
            pattern = r'\d\.\dV'
            info['voltage'] = re.search(pattern, info['ICMT'])[0]
            # Extract the temperature.
            pattern = r'-?\d\d?\.\dC'
            info['temperature'] = re.search(pattern, info['ICMT'])[0]
            # Extract the gain setting.
            pattern = r'(\w+-?\w*) gain'
            info['gain'] = re.search(pattern, info['ICMT'])[1]
        elif data == b'IART':
            # This is the artist subchunk. Store in a string.
            size = struct.unpack('<I', file.read(4))[0]
            info['IART'] = (
                file.read(size).decode('ascii').rstrip('\x00'))
            # Extract the serial number
            pattern = r'AudioMoth ([0-9A-Z]+)'
            info['serial number'] = re.search(pattern, info['IART'])[1]
        data = file.read(4)
    # We are now at the data chunk.
    # Nsamples is the number samples in the file (should be total time * fs).
    info['Nsamples'] = struct.unpack('<I', file.read(4))[0] // \
        (info['block_align'])
    # data0 is the position (in bytes) of the first data sample.
    info['data0'] = file.tell()
    info['wavetype'] = 'AudioMoth'
    return(info)
    